@@ -1,6 +1,6 @@
-/* $Id: DBIXS.h,v 11.2 2001/08/24 22:10:44 timbo Exp $
+/* $Id: DBIXS.h 10899 2008-03-10 12:00:51Z timbo $
  *
- * Copyright (c) 1994, 1995, 1996, 1997, 1998, 1999  Tim Bunce  England
+ * Copyright (c) 1994-2002  Tim Bunce  Ireland
  *
  * See COPYRIGHT section in DBI.pm for usage and distribution rights.
  */
@@ -23,6 +23,13 @@
 #undef debug
 #endif
 
+#ifdef std             /* causes problems with STLport <tscheresky@micron.com> */
+#undef std
+#endif
+
+/* define DBIXS_REVISION */
+#include "dbixs_rev.h"
+
 /* Perl backwards compatibility definitions */
 #include "dbipport.h"
 
@@ -37,6 +44,7 @@
  * and learns from the needs of various drivers.  See also the
  * DBISTATE_VERSION macro below. You can think of DBIXS_VERSION as
  * being a compile time check and DBISTATE_VERSION as a runtime check.
+ * By contract, DBIXS_REVISION is a driver source compatibility tool.
  */
 #define DBIXS_VERSION 93
 
@@ -49,28 +57,26 @@ error You_need_to_upgrade_your_DBI_module_before_building_this_driver
 #endif
 
 
-#if defined(USE_THREADS) && !defined(DBI_NO_THREADS)
-#define DBI_USE_THREADS
-#define DBI_LOCK	MUTEX_LOCK(DBIS->mutex)
-#define DBI_UNLOCK	MUTEX_UNLOCK(DBIS->mutex)
-#define dbi_mutex perl_mutex
-#define dbi_cond  perl_cond
-#else
 #define DBI_LOCK
 #define DBI_UNLOCK
-#define dbi_mutex void
-#define dbi_cond  void
-#endif
+
+#ifndef DBI_NO_THREADS
+#ifdef USE_ITHREADS
+#define DBI_USE_THREADS
+#endif /* USE_ITHREADS */
+#endif /* DBI_NO_THREADS */
 
 
-/* forward declaration of 'DBI Handle Common Data', see below		*/
+/* forward struct declarations						*/
 
+typedef struct dbistate_st dbistate_t;
 /* implementor needs to define actual struct { dbih_??c_t com; ... }*/
 typedef struct imp_drh_st imp_drh_t;	/* driver			*/
 typedef struct imp_dbh_st imp_dbh_t;	/* database			*/
 typedef struct imp_sth_st imp_sth_t;	/* statement			*/
 typedef struct imp_fdh_st imp_fdh_t;	/* field descriptor		*/
 typedef struct imp_xxh_st imp_xxh_t;	/* any (defined below)		*/
+#define DBI_imp_data_ imp_xxh_t		/* friendly for take_imp_data	*/
 
 
 
@@ -91,25 +97,25 @@ typedef struct dbih_com_std_st {
     HV   *my_h;		/* copy of outer handle HV (not refcounted)	*/
     SV   *parent_h;	/* parent inner handle (ref to hv) (r.c.inc)	*/
     imp_xxh_t *parent_com;	/* parent com struct shortcut		*/
-    dbi_cond  *thr_cond;/* condition for thread access (see dispatch)	*/
+    PerlInterpreter * thr_user;  /* thread that owns the handle         */
 
     HV   *imp_stash;	/* who is the implementor for this handle	*/
     SV   *imp_data;	/* optional implementors data (for perl imp's)	*/
 
     I32  kids;		/* count of db's for dr's, st's for db's etc	*/
     I32  active_kids;	/* kids which are currently DBIc_ACTIVE		*/
-    U32  thr_user;	/* thread id currently using the handle		*/
-    void *spare;
+    U32 pad;		/* keep binary compat */
+    dbistate_t *dbistate;
 } dbih_com_std_t;
 
 typedef struct dbih_com_attr_st {
     /* These are copies of the Hash values (ref.cnt.inc'd)	*/
     /* Many of the hash values are themselves references	*/
-    SV *Debug;
+    SV *TraceLevel;
     SV *State;		/* Standard SQLSTATE, 5 char string	*/
     SV *Err;		/* Native engine error code		*/
     SV *Errstr;		/* Native engine error message		*/
-    SV *Handlers;
+    UV ErrCount;
     U32  LongReadLen;	/* auto read length for long/blob types	*/
     SV *FetchHashKeyName;	/* for fetchrow_hashref		*/
     /* (NEW FIELDS?... DON'T FORGET TO UPDATE dbih_clearcom()!)	*/
@@ -131,13 +137,13 @@ struct imp_xxh_st { struct dbih_com_st com; };
 typedef struct {		/* -- DRIVER --				*/
     dbih_com_std_t	std;
     dbih_com_attr_t	attr;
-    HV          *cached_kids;	/* $drh->connect_cached(...)		*/
+    HV          *_old_cached_kids; /* not used, here for binary compat */
 } dbih_drc_t;
 
 typedef struct {		/* -- DATABASE --			*/
     dbih_com_std_t	std;	/* \__ standard structure		*/
     dbih_com_attr_t	attr;	/* /   plus... (nothing else right now)	*/
-    HV          *cached_kids;	/* $dbh->prepare_cached(...)		*/
+    HV          *_old_cached_kids; /* not used, here for binary compat */
 } dbih_dbc_t;
 
 typedef struct {		/* -- STATEMENT --			*/
@@ -190,22 +196,47 @@ typedef struct {		/* -- FIELD DESCRIPTOR --		*/
 #define DBIc_THR_USER_NONE  	(0xFFFF)
 #define DBIc_IMP_STASH(imp)  	_imp2com(imp, std.imp_stash)
 #define DBIc_IMP_DATA(imp)  	_imp2com(imp, std.imp_data)
+#define DBIc_DBISTATE(imp)  	_imp2com(imp, std.dbistate)
+#define DBIc_LOGPIO(imp)  	DBIc_DBISTATE(imp)->logfp
 #define DBIc_KIDS(imp)  	_imp2com(imp, std.kids)
 #define DBIc_ACTIVE_KIDS(imp)  	_imp2com(imp, std.active_kids)
+#define DBIc_LAST_METHOD(imp)  	_imp2com(imp, std.last_method)
 
-#define DBIc_DEBUG(imp)		(_imp2com(imp, attr.Debug))
-#define DBIc_DEBUGIV(imp)	SvIV(DBIc_DEBUG(imp))
+#define DBIc_TRACE_LEVEL_MASK	0x0000000F
+#define DBIc_TRACE_FLAGS_MASK	0xFFFFFF00
+#define DBIc_TRACE_SETTINGS(imp) (DBIc_DBISTATE(imp)->debug)
+#define DBIc_TRACE_LEVEL(imp)	(DBIc_TRACE_SETTINGS(imp) & DBIc_TRACE_LEVEL_MASK)
+#define DBIc_TRACE_FLAGS(imp)	(DBIc_TRACE_SETTINGS(imp) & DBIc_TRACE_FLAGS_MASK)
+/* DBIc_TRACE_MATCHES(this, crnt): true if this 'matches' (is within) crnt
+   DBIc_TRACE_MATCHES(foo, DBIc_TRACE_SETTINGS(imp))
+*/
+#define DBIc_TRACE_MATCHES(this, crnt)	\
+	(  ((crnt & DBIc_TRACE_LEVEL_MASK) >= (this & DBIc_TRACE_LEVEL_MASK)) \
+	|| ((crnt & DBIc_TRACE_FLAGS_MASK)  & (this & DBIc_TRACE_FLAGS_MASK)) )
+/* DBIc_TRACE: true if flags match & DBI level>=flaglevel, or if DBI level>level
+   This is the main trace testing macro to be used by drivers.
+   (Drivers should define their own DBDtf_* macros for the top 8 bits: 0xFF000000)
+   DBIc_TRACE(imp,         0, 0, 4) = if level >= 4
+   DBIc_TRACE(imp, DBDtf_FOO, 2, 4) = if tracing DBDtf_FOO & level>=2 or level>=4
+   DBIc_TRACE(imp, DBDtf_FOO, 2, 0) = as above but never trace just due to level
+*/
+#define DBIc_TRACE(imp, flags, flaglevel, level)	\
+	(  (flags && (DBIc_TRACE_FLAGS(imp) & flags) && (DBIc_TRACE_LEVEL(imp) >= flaglevel)) \
+	|| (level && DBIc_TRACE_LEVEL(imp) >= level) )
+
+#define DBIc_DEBUG(imp)		(_imp2com(imp, attr.TraceLevel)) /* deprecated */
+#define DBIc_DEBUGIV(imp)	SvIV(DBIc_DEBUG(imp))		 /* deprecated */
 #define DBIc_STATE(imp)		SvRV(_imp2com(imp, attr.State))
 #define DBIc_ERR(imp)		SvRV(_imp2com(imp, attr.Err))
 #define DBIc_ERRSTR(imp)	SvRV(_imp2com(imp, attr.Errstr))
-#define DBIc_HANDLERS(imp)	SvRV(_imp2com(imp, attr.Handlers))
+#define DBIc_ErrCount(imp)	_imp2com(imp, attr.ErrCount)
 #define DBIc_LongReadLen(imp)  	_imp2com(imp, attr.LongReadLen)
 #define DBIc_LongReadLen_init	80	/* may change */
 #define DBIc_FetchHashKeyName(imp) (_imp2com(imp, attr.FetchHashKeyName))
 
 /* handle sub-type specific fields						*/
-/*	dbh	*/
-#define DBIc_CACHED_KIDS(imp)  	_imp2com(imp, cached_kids)
+/*	dbh & drh	*/
+#define DBIc_CACHED_KIDS(imp)  	Nullhv /* no longer used, here for src compat */
 /*	sth	*/
 #define DBIc_NUM_FIELDS(imp)  	_imp2com(imp, num_fields)
 #define DBIc_NUM_PARAMS(imp)  	_imp2com(imp, num_params)
@@ -215,28 +246,34 @@ typedef struct {		/* -- FIELD DESCRIPTOR --		*/
 #define DBIc_FDESC_AV(imp)  	_imp2com(imp, fields_fdav)
 #define DBIc_FDESC(imp, i)  	((imp_fdh_t*)(void*)SvPVX(AvARRAY(DBIc_FDESC_AV(imp))[i]))
 
+/* XXX --- DO NOT CHANGE THESE VALUES AS THEY ARE COMPILED INTO DRIVERS --- XXX */
 #define DBIcf_COMSET	  0x000001	/* needs to be clear'd before free'd	*/
 #define DBIcf_IMPSET	  0x000002	/* has implementor data to be clear'd	*/
 #define DBIcf_ACTIVE	  0x000004	/* needs finish/disconnect before clear	*/
 #define DBIcf_IADESTROY	  0x000008	/* do DBIc_ACTIVE_off before DESTROY	*/
 #define DBIcf_WARN  	  0x000010	/* warn about poor practice etc  	*/
 #define DBIcf_COMPAT  	  0x000020	/* compat/emulation mode (eg oraperl)	*/
-
 #define DBIcf_ChopBlanks  0x000040	/* rtrim spaces from fetch char columns	*/
 #define DBIcf_RaiseError  0x000080	/* throw exception (croak) on error	*/
 #define DBIcf_PrintError  0x000100	/* warn() on error			*/
 #define DBIcf_AutoCommit  0x000200	/* dbh only. used by drivers		*/
 #define DBIcf_LongTruncOk 0x000400	/* truncation to LongReadLen is okay	*/
 #define DBIcf_MultiThread 0x000800	/* allow multiple threads to enter	*/
-#define DBIcf_Taint       0x001000	/* taint fetched data			*/
-#define DBIcf_ShowErrorStatement  0x002000	/* include Statement in error	*/
+#define DBIcf_HandleSetErr 0x001000	/* has coderef HandleSetErr attribute	*/
+#define DBIcf_ShowErrorStatement  0x002000   /* include Statement in error	*/
 #define DBIcf_BegunWork   0x004000	/* between begin_work & commit/rollback */
-#define DBIcf_HandleError 0x008000	/* has coderef in HandleError attribute	*/
+#define DBIcf_HandleError 0x008000	/* has coderef in HandleError attribute */
+#define DBIcf_Profile     0x010000	/* profile activity on this handle      */
+#define DBIcf_TaintIn     0x020000	/* check inputs for taintedness */
+#define DBIcf_TaintOut    0x040000	/* taint outgoing data */
+#define DBIcf_Executed    0x080000	/* do/execute called since commit/rollb */
+#define DBIcf_PrintWarn   0x100000	/* warn() on warning (err="0")		*/
+#define DBIcf_Callbacks   0x200000	/* has Callbacks attribute hash		*/
+/* NOTE: new flags may require clone() to be updated */
 
 #define DBIcf_INHERITMASK		/* what NOT to pass on to children */	\
   (U32)( DBIcf_COMSET | DBIcf_IMPSET | DBIcf_ACTIVE | DBIcf_IADESTROY		\
-  /* These are for dbh only:	*/						\
-  | DBIcf_AutoCommit | DBIcf_BegunWork	)
+  | DBIcf_AutoCommit | DBIcf_BegunWork | DBIcf_Executed | DBIcf_Callbacks )
 
 /* general purpose bit setting and testing macros			*/
 #define DBIbf_is( bitset,flag)		((bitset) &   (flag))
@@ -266,7 +303,7 @@ typedef struct {		/* -- FIELD DESCRIPTOR --		*/
 	imp_xxh_t *ph_com = DBIc_PARENT_COM(imp);			\
 	if (!DBIc_ACTIVE(imp) && ph_com && !dirty			\
 		&& ++DBIc_ACTIVE_KIDS(ph_com) > DBIc_KIDS(ph_com))	\
-	    croak("panic: DBI active kids (%d) > kids (%d)",		\
+	    croak("panic: DBI active kids (%ld) > kids (%ld)",		\
 		DBIc_ACTIVE_KIDS(ph_com), DBIc_KIDS(ph_com));		\
 	DBIc_FLAGS(imp) |=  DBIcf_ACTIVE;				\
     } while(0)
@@ -276,7 +313,7 @@ typedef struct {		/* -- FIELD DESCRIPTOR --		*/
 	if (DBIc_ACTIVE(imp) && ph_com && !dirty			\
 		&& (--DBIc_ACTIVE_KIDS(ph_com) > DBIc_KIDS(ph_com)	\
 		   || DBIc_ACTIVE_KIDS(ph_com) < 0) )			\
-	    croak("panic: DBI active kids (%d) < 0 or > kids (%d)",	\
+	    croak("panic: DBI active kids (%ld) < 0 or > kids (%ld)",	\
 		DBIc_ACTIVE_KIDS(ph_com), DBIc_KIDS(ph_com));		\
 	DBIc_FLAGS(imp) &= ~DBIcf_ACTIVE;				\
     } while(0)
@@ -295,7 +332,7 @@ typedef struct {		/* -- FIELD DESCRIPTOR --		*/
 
 
 #ifdef IN_DBI_XS		/* get Handle Common Data Structure	*/
-#define DBIh_COM(h)         	(dbih_getcom(h))
+#define DBIh_COM(h)         	(dbih_getcom2(aTHX_ h, 0))
 #else
 #define DBIh_COM(h)         	(DBIS->getcom(h))
 #define neatsvpv(sv,len)       	(DBIS->neat_svpv(sv,len))
@@ -318,15 +355,13 @@ typedef struct {		/* -- FIELD DESCRIPTOR --		*/
 #define DBI_IMP_SIZE(n,s) sv_setiv(perl_get_sv((n), GV_ADDMULTI), (s)) /* XXX */
 
 
-/* --- Implementors Field Descriptor Support --- */
-
 
 /* --- Event Support (VERY LIABLE TO CHANGE) --- */
 
-#define DBIh_EVENTx(h,t,a1,a2)	(DBIS->event((h), (t), (a1), (a2)))
-#define DBIh_EVENT0(h,t)	DBIh_EVENTx((h), (t), &sv_undef, &sv_undef)
-#define DBIh_EVENT1(h,t, a1)	DBIh_EVENTx((h), (t), (a1),      &sv_undef)
-#define DBIh_EVENT2(h,t, a1,a2)	DBIh_EVENTx((h), (t), (a1),      (a2))
+#define DBIh_EVENTx(h,t,a1,a2)	/* deprecated XXX */ &PL_sv_no
+#define DBIh_EVENT0(h,t)	DBIh_EVENTx((h), (t), &PL_sv_undef, &PL_sv_undef)
+#define DBIh_EVENT1(h,t, a1)	DBIh_EVENTx((h), (t), (a1),         &PL_sv_undef)
+#define DBIh_EVENT2(h,t, a1,a2)	DBIh_EVENTx((h), (t), (a1),         (a2))
 
 #define ERROR_event	"ERROR"
 #define WARN_event	"WARN"
@@ -334,24 +369,29 @@ typedef struct {		/* -- FIELD DESCRIPTOR --		*/
 #define DBEVENT_event	"DBEVENT"
 #define UNKNOWN_event	"UNKNOWN"
 
+#define DBIh_SET_ERR_SV(h,i, err, errstr, state, method) \
+	(DBIc_DBISTATE(i)->set_err_sv(h,i, err, errstr, state, method))
+#define DBIh_SET_ERR_CHAR(h,i, err_c, err_i, errstr, state, method) \
+	(DBIc_DBISTATE(i)->set_err_char(h,i, err_c, err_i, errstr, state, method))
+
 
 /* --- Handy Macros --- */
 
 #define DBIh_CLEAR_ERROR(imp_xxh) (void)( \
 	(void)SvOK_off(DBIc_ERR(imp_xxh)),    	\
 	(void)SvOK_off(DBIc_ERRSTR(imp_xxh)),	\
-	(SvPOK(DBIc_STATE(imp_xxh)) ? SvCUR(DBIc_STATE(imp_xxh))=0 : 0)	\
+	(void)SvOK_off(DBIc_STATE(imp_xxh))	\
     )
 
 
 /* --- DBI State Structure --- */
 
-typedef struct {
+struct dbistate_st {
 
 #define DBISTATE_VERSION  94	/* Must change whenever dbistate_t does	*/
 
     /* this must be the first member in structure			*/
-    void (*check_version) _((char *name,
+    void (*check_version) _((const char *name,
 		int dbis_cv, int dbis_cs, int need_dbixs_cv,
 		int drc_s, int dbc_s, int stc_s, int fdc_s));
 
@@ -368,22 +408,30 @@ typedef struct {
     char      * (*neat_svpv)	_((SV *sv, STRLEN maxlen));
     imp_xxh_t * (*getcom)	_((SV *h));	/* see DBIh_COM macro	*/
     void        (*clearcom)	_((imp_xxh_t *imp_xxh));
-    SV        * (*event)	_((SV *h, char *name, SV*, SV*));
+    SV        * (*event)	_((SV *h, const char *name, SV*, SV*));
     int         (*set_attr_k)	_((SV *h, SV *keysv, int dbikey, SV *valuesv));
     SV        * (*get_attr_k)	_((SV *h, SV *keysv, int dbikey));
     AV        * (*get_fbav)	_((imp_sth_t *imp_sth));
-    SV        * (*make_fdsv)	_((SV *sth, char *imp_class, STRLEN imp_size, char *col_name));
+    SV        * (*make_fdsv)	_((SV *sth, const char *imp_class, STRLEN imp_size, const char *col_name));
     int         (*bind_as_num)	_((int sql_type, int p, int s, int *t, void *v));
-    int         (*hash)		_((char *string, long i));
+    I32         (*hash)		_((const char *string, long i));
     SV        * (*preparse)	_((SV *sth, char *statement, IV ps_return, IV ps_accept, void *foo));
 
-    SV *neatsvpvlen;	/* only show dbgpvlen chars when debugging pv's	*/
+    SV *neatsvpvlen;		/* only show dbgpvlen chars when debugging pv's	*/
 
-    dbi_mutex	*mutex;
+    PerlInterpreter * thr_owner;	/* thread that owns this dbistate	*/
 
-    int         (*logmsg)	_((imp_xxh_t *imp_xxh, char *fmt, ...));
-    void *pad[8];
-} dbistate_t;
+    int         (*logmsg)	_((imp_xxh_t *imp_xxh, const char *fmt, ...));
+    int         (*set_err_sv)	_((SV *h, imp_xxh_t *imp_xxh, SV   *err, SV   *errstr, SV   *state, SV   *method));
+    int         (*set_err_char) _((SV *h, imp_xxh_t *imp_xxh, const char *err, IV err_i, const char *errstr, const char *state, const char *method));
+    int         (*bind_col)     _((SV *sth, SV *col, SV *ref, SV *attribs));
+
+    IO *logfp_ref;	/* DAA keep ptr to filehandle for refcounting */
+
+    /* WARNING: Only add new structure members here, and reduce pad2 to keep */
+    /* the memory footprint exactly the same */
+    void *pad2[4];
+};
 
 /* macros for backwards compatibility */
 #define set_attr(h, k, v)	set_attr_k(h, k, 0, v)
@@ -403,25 +451,23 @@ typedef struct {
 /* note that USE_ITHREADS implies MULTIPLICITY                      */
 #if defined(MULTIPLICITY) || defined(PERL_OBJECT) || defined(PERL_CAPI)
 
-# define DBISTATE_DECLARE
+# define DBISTATE_DECLARE typedef int dummy_dbistate /* keep semicolon from feeling lonely */
 # define DBISTATE_ASSIGN(st)
 # define DBISTATE_INIT
-static dbistate_t **get_dbistate() {
-    return ((dbistate_t**)&SvIVX(DBISTATE_ADDRSV));
-}
 # undef DBIS
-# define DBIS (*get_dbistate())
-# define dbis (*get_dbistate()) /* temp for bad drivers using 'dbis' instead of 'DBIS' */
+# define DBIS (*(INT2PTR(dbistate_t**, &SvIVX(DBISTATE_ADDRSV))))
+/* 'dbis' is temp for bad drivers using 'dbis' instead of 'DBIS' */
+# define dbis (*(INT2PTR(dbistate_t**, &SvIVX(DBISTATE_ADDRSV))))
 
 #else	/* plain and simple non perl object / multiplicity case */
 
 # define DBISTATE_DECLARE	static dbistate_t *DBIS
 # define DBISTATE_ASSIGN(st)	(DBIS = (st))
-# define DBISTATE_INIT_DBIS	DBISTATE_ASSIGN((dbistate_t*)SvIV(DBISTATE_ADDRSV))
+# define DBISTATE_INIT_DBIS	DBISTATE_ASSIGN(INT2PTR(dbistate_t*, SvIV(DBISTATE_ADDRSV)))
 # define DBISTATE_INIT {	/* typically use in BOOT: of XS file	*/    \
     DBISTATE_INIT_DBIS;	\
     if (DBIS == NULL)	\
-	croak("Unable to get DBI state. DBI not loaded.");	\
+	croak("Unable to get DBI state from %s at %p. DBI not loaded.", DBISTATE_PERLNAME, (void*)DBISTATE_ADDRSV); \
     DBIS->check_version(__FILE__, DBISTATE_VERSION, sizeof(*DBIS), NEED_DBIXS_VERSION, \
 		sizeof(dbih_drc_t), sizeof(dbih_dbc_t), sizeof(dbih_stc_t), sizeof(dbih_fdc_t) \
     ); \
@@ -441,21 +487,32 @@ static dbistate_t **get_dbistate() {
     if ((attribs) && SvOK(attribs)) {		\
 	if (!SvROK(attribs) || SvTYPE(SvRV(attribs))!=SVt_PVHV)		\
 	    croak("%s->%s(...): attribute parameter '%s' is not a hash ref",	\
-		    SvPV(h,na), func, SvPV(attribs,na));		\
+		    SvPV_nolen(h), func, SvPV_nolen(attribs));		\
     } else (attribs) = Nullsv
 
 #define DBD_ATTRIB_GET_SVP(attribs, key,klen)			\
 	(DBD_ATTRIB_OK(attribs)					\
 	    ? hv_fetch((HV*)SvRV(attribs), key,klen, 0)		\
 	    : (SV **)Nullsv)
-	
-#define DBD_ATTRIB_GET_BOOL(attribs, key,klen, svp, var)		\
-	if ((svp=DBD_ATTRIB_GET_SVP(attribs, key,klen)) != NULL)	\
-	    var = SvTRUE(*svp)
 
 #define DBD_ATTRIB_GET_IV(attribs, key,klen, svp, var)			\
 	if ((svp=DBD_ATTRIB_GET_SVP(attribs, key,klen)) != NULL)	\
 	    var = SvIV(*svp)
 
+#define DBD_ATTRIB_GET_BOOL(attribs, key,klen, svp, var)		\
+	if ((svp=DBD_ATTRIB_GET_SVP(attribs, key,klen)) != NULL)	\
+	    var = SvTRUE(*svp)
+
+#define DBD_ATTRIB_TRUE(attribs, key,klen, svp)				\
+	(  ((svp=DBD_ATTRIB_GET_SVP(attribs, key,klen)) != NULL)	\
+	    ? SvTRUE(*svp) : 0 )
+
+#define DBD_ATTRIB_GET_PV(attribs, key,klen, svp, dflt)			\
+	(((svp=DBD_ATTRIB_GET_SVP(attribs, key,klen)) != NULL)		\
+	    ? SvPV_nolen(*svp) : (dflt))
+
+#define DBD_ATTRIB_DELETE(attribs, key, klen)			\
+	hv_delete((HV*)attribs, key, klen, G_DISCARD)
+
 #endif /* DBIXS_VERSION */
 /* end of DBIXS.h */
